# include "std.h"
# if LICENCE
   This file is part of cwrite.
   Copyright 2007, Desislav Georgiev.
 
   cwrite is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   cwrite is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with cwrite.  If not, see <http://www.gnu.org/licenses/>.
# endif
# if HEADER
   # include <stdlib.h>

   # include "cwrite.h"
   # include "dataline.h"
   # include "datasystem.h"
   # include "status_line.h"
   # include "string.h"
# endif
# if GLOBAL
   char c_block_denominator_begin = '{';
   char c_block_denominator_end = '}';

   static bool block_error = 0;

   int block_count;
   int block_level;

# endif

int  c_block_header(Datachain * _datachain)
   { if ( ! _datachain->size )
        { return 0;
        }

     char * _dataline_point = _datachain->_line;
     _dataline_point += _datachain->size - 1;
     if ( * _dataline_point == c_block_denominator_begin )
        { return 1;
        }
     Datachain * _datachain_next = _datachain->_next;
     if ( ! _datachain_next )
        { return 0;
        }
     _dataline_point = (char *) _dataline_advance(_datachain_next, _datachain_next->_line);
     if ( ! _dataline_point )
        { return 0;
        }
     if ( * _dataline_point == c_block_denominator_begin )
        { return 2;
        }
     return 0;
   }
bool c_block_begin(Datachain * _datachain)
   { char * _dataline_point = _datachain->_line;
     int i, k;
     bool block_found = 0;

     for ( i = 0; i < _datachain->size; i++ )
        { if ( * _dataline_point++ == c_block_denominator_begin )
             { if ( * _dataline_point == '\'' )
                  { continue;
                  }
               if ( * _dataline_point == '\"' )
                  { continue;
                  }
               block_found = 1;
               break;
             }
        }
     if ( ! block_found )
        { return 0;
        }
     for ( k = i; k < _datachain->size; k++ )
        { if ( * _dataline_point++ == c_block_denominator_begin )
             { return 0;
             }
        }
     return 1;
   }
bool c_block_end(Datachain * _datachain)
   { char * _dataline_point = _datachain->_line;
     int i, k;
     bool block_found = 0;

     for ( i = 0; i < _datachain->size; i++ )
        { if ( * _dataline_point++ == c_block_denominator_end )
             { block_found = 1;
               break;
             }
        }
     if ( ! block_found )
        { return 0;
        }
     for ( k = i; k < _datachain->size; k++ )
        { if ( * _dataline_point++ == c_block_denominator_end )
            { return 0;
            }
        }
     return 1;
   }

char * _c_preprocessor_line(Datachain * _datachain)
   { int sid = 102;

     char * _dataline_point = _dataline_advance(_datachain, _datachain->_line);

     if ( ! _dataline_point )
        { //fail(sid, "no");
          return 0;
        }
     if ( * _dataline_point != '#' ) 
         { //stop(sid, "no2"); printf("%c\n", * _dataline_point); exit(-1);
           return 0;
         }
     //fail(sid, "yes");
     _dataline_point = _dataline_advance(_datachain, ++_dataline_point);
     //stop(sid, "last"); printf("%c\n", * _dataline_point); exit(-1);
     return _dataline_point;
   }

void c_preprocessor_block_begin(Datachain * _datachain)
   { char * _if     = (char *) "if";
     char * _ifdef  = (char *) "ifdef";
     char * _ifndef = (char *) "ifndef";
     char * _else   = (char *) "else";

     char * _dataline_point = _c_preprocessor_line(_datachain);
     if ( ! _dataline_point )
        { return;
        }

     if ( word_match(_dataline_point, _if) )
        { block_count++;
          block_level++;

          return;
        }
     if ( word_match(_dataline_point, _ifdef) || word_match(_dataline_point, _ifndef) )
        { block_count++;
          block_level++;

          return;
        }
     if ( word_match(_dataline_point, _else) )
        { block_count++;
          block_level++;

          return;
        }
     return;
   }
void c_preprocessor_block_end(Datachain * _datachain)
   { char * _else  = (char *) "else";
     char * _endif = (char *) "endif";

     char * _dataline_point = _c_preprocessor_line(_datachain);
     if ( _dataline_point && word_match(_dataline_point, _endif) )
        { block_count--;
        }

     if ( ! block_count )
        { return;
        }

     char * _dataline_point_next;
     if ( _datachain->_next )
        { _dataline_point_next = _c_preprocessor_line(_datachain->_next);
          if ( _dataline_point_next && word_match(_dataline_point_next, _else) )
             { block_count--;
             }
       }

     return;
   }

int  block_locate_c_common(Datachain * _datachain_block_header, Datachain ** __datachain_block_begin, Datachain ** __datachain_block_end)
   { int block_count = 0;
     int block_level = 0;

     if ( _datachain_block_header == * __datachain_last )
        { return 0;
        }
     int block_header = c_block_header(_datachain_block_header);
     if ( ! block_header )
        { return 0;
        }
     if ( block_header == 1 )
        { block_count = 1;
          block_level = 1;
        }
     Datachain * _datachain = _datachain_block_header->_next;
     * __datachain_block_begin = _datachain;
    
     while ( _datachain )
        { if ( c_block_begin(_datachain) )
             { block_count++;
               block_level++;
             }
          if ( c_block_end(_datachain) )
             { block_count--;
             }
          if ( ! block_count )
             { * __datachain_block_end = _datachain;
               return block_level;
             }
          _datachain = _datachain->_next;
        }    
     return -1;
   }
int  block_locate_c_preprocessor(Datachain * _datachain_block_header, Datachain ** __datachain_block_begin, Datachain ** __datachain_block_end)
   { block_count = 0;
     block_level = 0;

        // -- last chain, no block --
     if ( _datachain_block_header == * __datachain_last )
        { return 0;
        }

     Datachain * _datachain = _datachain_block_header;
     * __datachain_block_begin = _datachain->_next;
     while ( _datachain )
        { c_preprocessor_block_begin(_datachain);
          if ( (_datachain == _datachain_block_header) && ! block_level )
             { return 0;
             }
          c_preprocessor_block_end(_datachain);

          if ( ! block_count )
             { * __datachain_block_end = _datachain;
               break;
             }
          _datachain = _datachain->_next;
       }
     if ( ! _datachain )
         { return -1;
         }
     return block_level;
   }
