# include "std.h"
# if LICENCE
   This file is part of cwrite.
   Copyright 2007, Desislav Georgiev.
 
   cwrite is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   cwrite is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with cwrite.  If not, see <http://www.gnu.org/licenses/>.
# endif
# if HEADER
   # include <stdlib.h>
   # include <crypt.h>
   # include <shadow.h>
   # include <fcntl.h>
   # include <unistd.h>
   # include <errno.h>

   # include "block.h"
   # include "color_theme.h"
   # include "command.h"
   # include "command_line.h"
   # include "cursor.h"
   # include "cwrite.h"
   # include "dataline.h"
   # include "datasystem.h"
   # include "file.h"
   # include "header_line.h"
   # include "history.h"
   # include "io.h"
   # include "message.h"
   # include "mycurses.h"
   # include "preferences.h"
   # include "status_line.h"
   # include "screen.h"
   # include "search.h"
   # include "string.h"
   # include "view.h"
# endif
# if PROTO
   void command_save();
   void command_close();

   void command_exit();
   void command_new();
   void command_open_0();
   void command_close();
   void command_save();
   void command_locate();
   void command_find();
   void command_replace_s();
   void command_help();
   void command_lock();
   void command_set();

   void command_set_color_theme();
   void command_set_tab();
   void command_set_syntax();
   void command_set_lf();
# endif
# if GLOBAL
   bool return_request = 0;
   char * _command[] = 
      { (char *) "exit",
        (char *) "new",
        (char *) "open",
        (char *) "close",
        (char *) "save",
        (char *) "locate",
        (char *) "find",
        (char *) "replace",
        (char *) "help",
        (char *) "lock",
        (char *) "set",
      };
   void (* _command_set_f[])() =
      { command_set_color_theme,
        command_set_syntax,
        command_set_tab,
        command_set_lf,
      };
   void (* _command_f[])() =
      { command_exit,
        command_new,
        command_open_0,
        command_close,
        command_save,
        command_locate,
        command_find,
        command_replace_s,
        command_help,
        command_lock,
        command_set,
      };  
# endif
# if TEST
   # define TEST_COMMAND_SEARCH
   //# define TEST_COMMAND_LOCK
# endif

int command_recognize()
   { for ( int i = 0; i < sizeof(_command)/sizeof(char *); i++ )
        { if ( string_compare(_command[i], _entry_line) )
             { return i;
             }
        }
     return -1;
   }

void command_exit()
   { int k;

     while ( datasystem_count )
        { datasystem_ordinal = DATASYSTEM_FIRST;
          datasystem_set();
          k = _datasystem->ordinal;
          if ( * _status_register & MODIFY_STATE )
             { header_line_refresh();
               screen_redraw();
             }
          command_close();
          if ( k == _datasystem->ordinal && * _status_register & MODIFY_STATE )
             { return;
             }
        }
     exit_request = 1;
     return;
   }
void command_new()
   { if ( datasystem_count == DATASYSTEM_CARDINAL )
        { status_line_print(_message[7]);
          return;  
        }
     file_new();
     return;
   }
void command_open(char * _filename)
   { if ( datasystem_count == DATASYSTEM_LAST )
        { status_line_print(_message[7]);
          return;
        }
     if ( ! _filename )
        { command_line_echo(_message[4]);
          command_line_read(_history_ptr[HISTORY_FILE_OPEN]);

         if ( ! (* _entry_line) )
            { return;
            }
        }
     else
        { int filename_len = string_len(_filename);
          if ( filename_len >= entry_line_memsize )
             { status_line_print(_message[76]);
               return;
             }
          string_copy(_entry_line, _filename);
        }

        // -- indicate directory --
     int fd = open(_entry_line, O_RDWR);
     if ( fd == -1 )
        { switch (errno)
             { case EISDIR:
                  status_line_print(_message[48]);
                  return;
             }
        }
     else
        { io_close(fd);
        }

        // -- check if file is open --
     int file_ordinal = file_open_ordinal(_entry_line);
     if ( file_ordinal >= DATASYSTEM_FIRST )
        { if ( file_ordinal != datasystem_ordinal )
             { datasystem_ordinal = file_ordinal;
               datasystem_set();
               header_line_refresh();
               screen_redraw();
             }
          status_line_print(_message[12]);

          return;
        }

        // -- open file --
     int open_event = file_open();
     switch(open_event)
        { case FILE_OPEN_FAILURE:
             if ( errno == EACCES )
                { status_line_print(_message[47]);
                }
             else
                { status_line_print(_message[13]);
                }
             return;
          case FILE_OPEN_CTRL_CHAR:
             status_line_print(_message[45]);
             return;
          case FILE_OPEN_BS_CHAR:
             status_line_print(_message[46]);
             return;
        }

     * __datachain = * __datachain_first;

        // -- redraw screen --
     header_line_refresh();
     status_line_refresh();
     screen_redraw();
     cursor_move(* _cursor_y, * _cursor_x);
     return;
   }
void command_close()
   { if ( * _status_register & MODIFY_STATE )
        { if ( * _datasystem->_filename )
             { command_line_echo(_message[5]);
             }
          else
             { command_line_echo(_message[6]);
             }
          int user_request = command_line_revolve(_confirm_str, 0, 2);
          command_line_clear();
          command_line_redraw();
          cursor_show();
          if ( user_request == REPLY_CANCEL )
            { return;
            }
          if ( user_request == REPLY_YES )
             { command_save();
               if ( * _status_register & MODIFY_STATE )
                  { return;
                  }
             }
        }
     file_close();
     
     return;
   }
void command_save_direct()
   { if ( ! (* _status_register & MODIFY_STATE) )
        { return;
        }
     if ( ! * _datasystem->_filename )
        { command_save();
          return;
        }
     file_save();
     //* _status_register &= ~MODIFY_STATE;
     return;
   }
void command_save()
   { if ( * _datasystem->_filename )
        { command_save_direct();
          return;
        }
     command_line_echo(_message[11]);
     command_line_read(_history_ptr[HISTORY_FILE_SAVE]);
     if ( ! (* _entry_line) )
        { return;
        }
     int fd = open(_entry_line, O_RDONLY);
     if ( fd != -1 )
        { io_close(fd);
          command_line_echo(_message[14]);
          int user_request = command_line_revolve(_confirm_str, 1, 2);
          if ( user_request == REPLY_CANCEL || user_request == REPLY_NO )
             { return;
             }
        }
     string_copy(_datasystem->_filename, _entry_line);
     if ( file_save() )
        { * _status_register &= ~MODIFY_STATE;
        }
     else
        { * _datasystem->_filename = 0;
        }
     header_line_refresh();

     return;
   }
void command_search()
   { if ( search_submode == SEARCH_START )
        { if ( search_mode == SEARCH_LOCATE )
            { command_line_echo(_message[73]);
            }
          else
            { command_line_echo(_message[74]);
            }
          command_line_read(_history_ptr[HISTORY_SEARCH]);
        }
     if ( ! (* _entry_line) )
        { return;
        }
     if ( search_submode == SEARCH_START )
        { string_copy(_find_str, _entry_line);
        }
     # ifdef TEST_COMMAND_SEARCH
        status_line_print(_find_str);
     # endif
     int match = search_execute();
     if ( match == -1 )
        { status_line_print(_message[33]);
        }
     if ( match == 0 )
        { status_line_print(_message[34]);
        }
     if ( match > 0 )
        { status_line_print(_message[35]);
        }
     return;
   }
void command_replace()
   { if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }

     search_mode = SEARCH_FIND;
     search_mode |= SEARCH_REPLACE;

        // -- read text to find --
     command_line_echo(_message[18]);
     command_line_read(_history_ptr[HISTORY_SEARCH]);
     if ( ! * _entry_line )
        { return;
        }
     string_copy(_find_str, _entry_line);

        // -- read text to replace --
     command_line_clear();
     command_line_print(_message[19], _find_str);
     command_line_stop = string_len(_message[19]) - 2 + string_len(_find_str);
     command_line_size = command_line_stop;
     command_line_redraw();
     command_line_read(_history_ptr[HISTORY_SEARCH]);
     if ( ! * _entry_line )
        { return;
        }
     string_copy(_replace_str, _entry_line);

     int match = search_execute();
     if ( match == -1 )
        { status_line_print(_message[33]);
        }
     if ( match == 0 )
        { status_line_print(_message[34]);
        }
     if ( match > 0 )
        { status_line_print(_message[35]);
        }
     return;
   }
void command_help()
   { command_open(_help_filename);
     _datasystem->read_only = 1;
     view_collapse_max();

     return;
   }
void command_lock()
   { 
     # ifdef LOCK_OPTION
        int key;
        do
           { curses_clear();
             cursor_hide();
             do
                { key = curses_getch();
                }
             while ( key );
             cursor_show();
             command_line_echo(_message[24]);
             command_line_read_lock();
           }
        while ( ! string_compare(_user_spwd->sp_pwdp, crypt(_entry_line, _user_spwd->sp_pwdp)) );
        // || ! string_compare(_user_data->pw_passwd, crypt(_entry_line, _user_data->pw_passwd)) );
        # ifdef TEST_COMMAND_LOCK
           while (0);
           endwin();
           printf("%s\n", _entry_line);
           printf("%s\n", _user_data->pw_passwd);
           printf("%s\n", crypt(_entry_line, _user_data->pw_passwd));
           //spwd * _shadow_pwd = getspent();
           printf("%s\n", _user_spwd->sp_pwdp);
           printf("%s\n", crypt(_entry_line, _user_spwd->sp_pwdp));
           exit(0);
        # endif
        header_line_redraw();
        status_line_redraw();
        if ( datasystem_count )
           { screen_redraw();
           }
        return;
     # else
        return;
     # endif
   }

void command_block_detach()
   { Datachain * _datachain_block_begin, * _datachain_block_end;

     int block_level = block_locate(* __datachain, & _datachain_block_begin, & _datachain_block_end);
     switch (block_level)
        { case BLOCK_NO_SYNTAX:
             status_line_print(_message[40]);
             return;
          case BLOCK_ERROR:
             status_line_print(_message[8]);
             return;
          case BLOCK_NONE:
             status_line_print(_message[9]);
             return;
        }

     * __datachain_detach_start = * __datachain;
     * __datachain_detach_end = _datachain_block_end;
     block_detach();

     return;
   }
void command_block_attach()
   { if ( ! (* _status_register & DETACH_STATE) )
        { status_line_print(_message[10]);
          //curses_getch();
          return;
        }
     block_attach();

     return;
   }
void command_block_collapse()
   { Datachain * _datachain_block_begin, * _datachain_block_end;

     int block_level = block_locate(* __datachain, & _datachain_block_begin, & _datachain_block_end);
     switch (block_level)
        { case BLOCK_NO_SYNTAX:
             status_line_print(_message[40]);
             return;
          case BLOCK_ERROR:
             status_line_print(_message[8]);
             return;
          case BLOCK_NONE:
             status_line_print(_message[9]);
             return;
        }

     if ( _datachain_block_begin->invisible )
        { return;
        }
     block_collapse(_datachain_block_begin, _datachain_block_end);

     return;
   }
void command_block_expand()
   { Datachain * _datachain_block_begin, * _datachain_block_end;

     int block_level = block_locate(* __datachain, & _datachain_block_begin, & _datachain_block_end);
     switch (block_level)
        { case BLOCK_NO_SYNTAX:
             status_line_print(_message[40]);
             return;
          case BLOCK_ERROR:
             status_line_print(_message[8]);
             return;
          case BLOCK_NONE:
             status_line_print(_message[9]);
             return;
        }

     block_expand(_datachain_block_begin, _datachain_block_end);

     return;
   }

void command_set_color_theme()
   { command_line_echo(_message[26]);

     int user_request = command_line_revolve(_color_theme_str, _preferences->color_theme, color_theme_cardinal);
     if ( user_request >= 0 )
        { color_theme_set(user_request);
        }
     return;
   }
void command_set_tab()
   { command_line_echo(_message[27]);
     command_line_stop_print((char *) "%d", _preferences->tab_value);
     command_line_read_number();
     if ( ! * _entry_line )
        { return;
        }
     int tab_value = atoi(_entry_line);
     if ( tab_value == -1 )
        { return;
        }
     if ( tab_value > tab_limit )
        { status_line_vprintf(_message[75], tab_limit);
          return;
        }
     _preferences->tab_value = tab_value;
     return;
   }
void command_set_syntax()
   { command_line_echo(_message[41]);

     int user_request = command_line_revolve(_syntax_str, _preferences->syntax, syntax_cardinal);
     if ( user_request >= 0 )
        { _preferences->syntax = user_request;
        }
     return;
   }
void command_set_lf()
   { command_line_echo(_message[77]);

     int user_request = command_line_revolve(_lf_str, _preferences->line_feed, lf_cardinal);
     if ( user_request >= 0 )
        { _preferences->line_feed = user_request;
        }
     return;
   }

void command_set()
   { command_line_echo(_message[25]);

     int user_request = command_line_revolve(_set_str, _preferences->ordinal, set_cardinal);
     _command_set_f[user_request]();
     if ( user_request >= 0 )
        { _preferences->ordinal = user_request;
        }
     return;
   }
      
void command_insert_line()
   { switch(_preferences->line_feed)
        { case LF_NORMAL:
             dataline_insert();
             break;
          case LF_SMART:
             dataline_insert_smart();
             break;
        }
     return;
   }

void command_open_0()
   { command_open(0);
     return;
   }
void command_replace_s()
   { search_submode = SEARCH_START;
     command_replace();
   }
void command_locate()
   { search_mode = SEARCH_LOCATE;
     search_submode = SEARCH_START;
     command_search();
     return;
   }
void command_find()
   { search_mode = SEARCH_FIND;
     search_submode = SEARCH_START;
     command_search();
     return;
   }

void command_execute(int n)
   { _command_f[n]();
   }
