# include "std.h"
# if LICENCE
   This file is part of cwrite.
   Copyright 2007, Desislav Georgiev.
 
   cwrite is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   cwrite is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with cwrite.  If not, see <http://www.gnu.org/licenses/>.
# endif
# if HEADER
   # include <stdlib.h>

   # include "block.h"
   # include "color_theme.h"
   # include "cursor.h"
   # include "cwrite.h"
   # include "datachain.h"
   # include "dataline.h"
   # include "datasystem.h"
   # include "header_line.h"
   # include "key.h"
   # include "message.h"
   # include "mycurses.h"
   # include "screen.h"
   # include "scroll.h"
   # include "status_line.h"
   # include "string.h"
# endif
# if CONST
   # define SELECTION_NONE  0
   # define SELECTION_LEFT  1
   # define SELECTION_RIGHT 2
   # define SELECTION_UP    3
   # define SELECTION_DOWN  4

   # define COPY_FLAG 0
   # define CUT_FLAG  1
# endif
# if GLOBAL
   Datachain * _datachain_selection_begin;
   Datachain * _datachain_selection_end;
   Datachain * _datachain_selection_min;
   Datachain * _datachain_selection_max;

   int selection_x_direction;
   int selection_y_direction;
   int selection_start;
   int selection_end;
# endif
# if PROTO
   void selection_cursor_left();
   void selection_cursor_right();
   void selection_cursor_up();
   void selection_cursor_down();
   void selection_cursor_home();
   void selection_cursor_end();

   void selection_move_left();
   void selection_move_right();
   void selection_move_up();
   void selection_move_down();

   void selection_update();
   void selection_fix();
   bool selection_process();
   void selection_color();

   void selection_get(int flag);
   void selection_paste();

   bool datachain_is_selected(Datachain * _datachain);
   void selection_clear();

   bool block_select(Datachain *, Datachain **, Datachain **);

   int selection_visible_lines();
   int selection_line_max();
   void dataline_strip(Datachain * _datachain);
# endif

void datasystem_selection_init()
   {    // -- init _datasystem_selection --
     int datasystem_ordinal_tmp = datasystem_ordinal;
     datasystem_ordinal = DATASYSTEM_SELECTION;
     int line_max = selection_line_max();
     if ( line_max < dataline_start_size )
        { line_max = dataline_start_size;
        }
     datasystem_init(line_max);
     if ( string_len(_user_selection) <= datasystem_filename_memsize )
        { string_copy(_datasystem->_filename, _user_selection);
        }

        // -- restore _datasystem --
     datasystem_ordinal = datasystem_ordinal_tmp;
     datasystem_set();

     return;
   }
void datasystem_selection_free()
   { Datasystem * _datasystem_selection = _datasystem_pointer[DATASYSTEM_SELECTION];
     free(_datasystem_selection->_filename);
     Datachain * _datachain = _datasystem_selection->_datachain_first;
     Datachain * _datachain_next;
     while ( _datachain )
        { _datachain_next = _datachain->_next;
          free(_datachain->_line);
          free(_datachain);
          _datachain = _datachain_next;
        }
     return;
   }
bool datasystem_selection_none()
   { Datasystem * _datasystem_selection = _datasystem_pointer[DATASYSTEM_SELECTION];
     Datachain * _datachain = _datasystem_selection->_datachain_first;
     if ( _datachain )
        { return 0;
        }
     return 1;
   }

void selection_mark()
   { Datachain * _datachain;
     int c;
     bool p = 0;

     if ( datasystem_ordinal == DATASYSTEM_SELECTION )
        { return;
        }

     //datasystem_selection_free();
     status_line_print(_message[15]);

     _datachain_selection_begin = * __datachain;
     _datachain_selection_end = * __datachain;
     selection_y_direction = SELECTION_NONE;
     selection_x_direction = SELECTION_NONE;
     selection_start = * _dataline_offset;
     selection_end = * _dataline_offset;
     do
        { c = curses_getch();
          _datachain = (* __datachain);
          if ( selection_start == selection_end )
             { //selection_y_direction = SELECTION_NONE;
               selection_x_direction = SELECTION_NONE;
               _datachain->selection_start = * _dataline_offset;
               _datachain->selection_end = * _dataline_offset;
             }
          if ( _datachain_selection_begin == _datachain_selection_end )
             { selection_y_direction = SELECTION_NONE;
             }
          switch (c)
             { case KEY_LEFT:
                  selection_cursor_left();
                  break;
               case KEY_RIGHT:
                  selection_cursor_right();
                  break;
               case KEY_UP:
                  selection_cursor_up();
                  break;
               case KEY_DOWN:
                  selection_cursor_down();
                  break;
               case KEY_HOME:
                  selection_cursor_home();
                  break;
               case KEY_END:
                  selection_cursor_end();
                  break;
               case CTRL_J:
               case CTRL_M:
               case KEY_F(1):
                  selection_fix();
                  cursor_hide();
                  p = selection_process();
                  c = 0;
                  break;
             }
          if ( ! p )
             { selection_color();
             }
        }
     while ( c );

     selection_clear();
     if ( ! p )
        { status_line_refresh();
        }
     screen_redraw();
     cursor_show();
     return;
   }
void selection_fix()
   { selection_update();
   }
void selection_update()
   { Datachain * _datachain;

     (* __datachain)->selection_start = selection_start;
     (* __datachain)->selection_end = selection_end;

     if ( selection_y_direction == SELECTION_DOWN )
        { _datachain_selection_min = _datachain_selection_begin;
          _datachain_selection_max = _datachain_selection_end;
        }
     else
        { _datachain_selection_min = _datachain_selection_end;
          _datachain_selection_max = _datachain_selection_begin;
        }

     _datachain = * __datachain_first;
     while ( _datachain != _datachain_selection_min )
        { _datachain->selection_start = 0;
          _datachain->selection_end = 0;
          _datachain = _datachain->_next;
        }
     _datachain = _datachain_selection_max->_next;
     while ( _datachain )
        { _datachain->selection_start = 0;
          _datachain->selection_end = 0;
          _datachain = _datachain->_next;
        }

     return;
   }
bool selection_process()
   { status_line_print(_message[16]);

     int c;
     do
        { c = curses_getch();
          if ( ! c )
             { return 0;
             }
          if ( ! (* _status_register & TYPE_MODE) && c != CTRL_C )
             { continue;
             }
          switch (c)
             { case CTRL_C:
                  selection_get(COPY_FLAG);
                  status_line_print(_message[21]);
                  return 1;
                  break;
               case CTRL_X:
                  selection_get(CUT_FLAG);
                  status_line_print(_message[22]);
                  return 1;
                  break;
               case KEY_RIGHT:
                  selection_move_right();
                  break;
               case KEY_LEFT:
                  selection_move_left();
                  break;
               case KEY_UP:
                  selection_move_up();
                  break;
               case KEY_DOWN:
                  selection_move_down();
                  break;
             }
        }
     while ( 1 );
   }
void selection_color()
   { int color_offset;
     int color_start_point, color_end_point;
     int selection_start, selection_end, delta;
     int selection_y = * _cursor_y;
     bool tree_view = 0;

     screen_redraw();

     Datachain * _datachain = _datachain_selection_end;
     Datachain * _datachain_next;
     do
        { if ( _datachain->invisible )
             { goto L1;
             }
          selection_start = _datachain->selection_start;
          selection_end = _datachain->selection_end;
          delta = selection_end - selection_start;

          color_start_point = selection_start - * _dataline_offset + * _cursor_x;
          //color_offset = min(delta, screen_right - screen_left);
          color_offset = delta;

          if ( color_start_point < screen_left )
             { color_offset = delta + color_start_point - screen_left;
               color_start_point = screen_left;
             }
          else
             { if ( delta > screen_right - screen_left )
                  { color_offset = screen_right - screen_left + 1;
                    color_end_point = selection_end - * _dataline_offset + * _cursor_x;
                    if ( color_end_point == screen_right )
                       { color_offset--;
                       }
                  }
             }
          _datachain_next = _datachain->_next;
          if ( _datachain_next && datachain_is_selected(_datachain_next) )
             { if ( _datachain_next->invisible )
                  { tree_view = 1;
                    if ( (* _status_register & DETACH_STATE) && (_datachain == * __datachain_detach_end) )
                       { tree_view = 0;
                       }
                  }
             }
          if ( tree_view )
             { curses_mvaddnstr(selection_y, 1, "+", 1);
               curses_mvchgat(selection_y, 1, 1, A_NORMAL, selection_color_pair, 0);
             }
          curses_mvchgat(selection_y, color_start_point, color_offset, A_NORMAL, selection_color_pair, 0);
          curses_touchline(stdscr, selection_y, 1);

   L1:    if ( _datachain == _datachain_selection_begin )
             { break;
             }
          if ( selection_y_direction == SELECTION_UP )
             { if ( ! _datachain->invisible )
                  { selection_y++;
                    if ( selection_y > screen_bottom )
                       { break;
                       }
                  }
               _datachain = _datachain->_next;
             }
          else
             { if ( ! _datachain->invisible )
                  { selection_y--;
                    if ( selection_y < screen_top )
                       { break;
                       }
                  }
               _datachain = _datachain->_prev;
             }
          tree_view = 0;
        }
     while ( _datachain );

     cursor_move(* _cursor_y, * _cursor_x);
     return;
   }

void selection_cursor_left()
   { Datachain * _datachain = * __datachain;
     int delta = selection_end - selection_start;
     bool m = 0;

     if ( * _cursor_x == screen_left )
        { m = scroll_left();
        }
     else
        { if ( * _cursor_x == screen_right && delta > screen_right - screen_left )
             { m = scroll_left();
             }
        }
     if ( ! m )
        { m = cursor_left();
        }
     if ( ! m )
        { return;
        }
     switch (selection_x_direction)
        { case SELECTION_NONE:
             selection_x_direction = SELECTION_LEFT;
             selection_start = * _dataline_offset;
             selection_end = * _dataline_offset + 1;
             break;
          case SELECTION_LEFT:
             selection_start--;
             break;
          case SELECTION_RIGHT:
             selection_end--;
             break;
        }

     _datachain->selection_start = selection_start;
     _datachain->selection_end = selection_end;

     return;
   }
void selection_cursor_right()
   { Datachain * _datachain = * __datachain;
     int delta = selection_end - selection_start;
     bool m = 0;

     if ( * _cursor_x == screen_right )
        { m = scroll_right();
        }
     else
        { if ( * _cursor_x == screen_left && delta > screen_right - screen_left )
             { m = scroll_right();
             }
        }
     if ( ! m )
        { m = cursor_right();
        }
     if ( ! m )
        { return;
        }
     switch (selection_x_direction)
        { case SELECTION_NONE:
             selection_x_direction = SELECTION_RIGHT;
             selection_start = * _dataline_offset - 1;
             selection_end = * _dataline_offset;
             break;
          case SELECTION_LEFT:
             selection_start++;
             break;
          case SELECTION_RIGHT:
             selection_end++;
             break;
        }

     _datachain->selection_start = selection_start;
     _datachain->selection_end = selection_end;

     return;
   }
void selection_cursor_up()
   { bool m = 0, update_rq = 1, move_rq = 1;

     Datachain * _datachain = * __datachain;
     Datachain * _datachain_prev = _datachain->_prev;
     if ( ! _datachain_prev )
        { return;
        }
     if ( * _status_register & DETACH_STATE && _datachain == * __datachain_detach_start )
        { return;
        }
     Datachain *_datachain_next = (* __datachain)->_next;

     if ( _datachain_next->invisible )
        { if ( ! datachain_is_selected(* __datachain) )
             { return;
             }
        }
         
     if ( selection_y_direction == SELECTION_NONE )
        { selection_y_direction = SELECTION_UP;
        }
     if ( _datachain_prev->invisible )
        {   // -- select --
          if ( selection_y_direction == SELECTION_UP)
             { _datachain = (Datachain *) _next_visible_line_up(* __datachain);
               Datachain * _datachain_block_begin, * _datachain_block_end;
               if (! block_select(_datachain, & _datachain_block_begin, & _datachain_block_end) )
                  { return;
                  }
             }
          else
             { //if ( _datachain_next->invisible )
                  { _datachain_selection_end = (* __datachain)->_prev;
                    update_rq = 0;
                    if ( datachain_is_selected(_datachain_next) )
                       { move_rq = 0;
                         update_rq = 1;
                       }
                    selection_update();
                  }
             }
        }
     else
        { if ( _datachain_next->invisible )
             { if ( selection_y_direction == SELECTION_DOWN )
                  { // ( _datachain_selection_max->invisible  )
                       { move_rq = 0;
                         update_rq = 1;
                       }
                    if ( _datachain_selection_end == * __datachain )
                       { move_rq = 1;
                       }
                    if ( _datachain_selection_end == (* __datachain)->_prev )
                       { move_rq = 1;
                         update_rq = 1;
                       }
                  }
             }
        }
        // -- move upwards --
     if ( move_rq )
        { if ( * _cursor_y == screen_top )
             { m = scroll_up();
             }
          if ( * _cursor_y == screen_bottom && selection_visible_lines() > screen_bottom - screen_top )
             { m = scroll_up();
             }
          if ( ! m )
             { m = cursor_up();
             }
          if ( ! m )
             { return;
             }
        }
     if ( update_rq )
        { _datachain_selection_end = * __datachain;
          selection_update();
        }
     return;
   }
void selection_cursor_down()
   { Datachain * _datachain = _datachain_selection_end;
     Datachain * _datachain_next = _datachain->_next;
     bool m = 0;

     if ( ! _datachain_next )
        { return;
        }
     if ( (* _status_register & DETACH_STATE) && (_datachain == * __datachain_detach_end) )
        { return;
        }

     if ( selection_y_direction == SELECTION_NONE )
        { selection_y_direction = SELECTION_DOWN;
        }
        // -- select --
     if ( _datachain_next->invisible )
        {    // -- select block --
          if ( selection_y_direction == SELECTION_DOWN )
             { if ( ! datachain_is_selected(* __datachain) )
                  { return;
                  }
               Datachain * _datachain_block_begin, * _datachain_block_end;
               if ( ! block_select(_datachain, & _datachain_block_begin, & _datachain_block_end) )
                  { return;
                  }
               _datachain_selection_end = _datachain_block_end;
               selection_update();
               goto L1;
             }
        }

        // -- move downwards --
     if ( * _cursor_y == screen_bottom )
        { m = scroll_down();
        }
     if ( * _cursor_y == screen_top && selection_visible_lines() > screen_bottom - screen_top )
        { m = scroll_down();
        }
     if ( ! m )
        { m = cursor_down();
        }
     if ( ! m )
        { return;
        }

        // -- select line --
     _datachain_selection_end = * __datachain;
     selection_update();

     L1:
     return;
   }
void selection_cursor_home()
   { Datachain * _datachain = * __datachain;
     
     if ( selection_x_direction == SELECTION_NONE )
        { selection_x_direction = SELECTION_LEFT;
          selection_end = * _dataline_offset;
             // -- enters SELECTION_LEFT --
        }
     cursor_home();
     if ( selection_x_direction == SELECTION_LEFT )
        { selection_start = 0;
        }
     if ( selection_x_direction == SELECTION_RIGHT )
        { selection_end = selection_start;
          selection_start = 0;
          selection_x_direction = SELECTION_LEFT;
         }

     _datachain->selection_start = selection_start;
     _datachain->selection_end = selection_end;

     return;
   }
void selection_cursor_end()
   { Datachain * _datachain = * __datachain;

     if ( selection_x_direction == SELECTION_NONE )
        { selection_x_direction = SELECTION_RIGHT;
          selection_start = * _dataline_offset;
              // -- enters SELECTION_RIGHT --
        }

     cursor_end();

     if ( selection_x_direction == SELECTION_LEFT )
        { selection_start = selection_end;
          selection_x_direction = SELECTION_RIGHT;
             // -- enters SELECTION_RIGHT --
        }
     if ( selection_x_direction = SELECTION_RIGHT )
        { selection_end = _datachain->size;
        }

     _datachain->selection_start = selection_start;
     _datachain->selection_end = selection_end;

     return;
   }

void selection_get(int flag)
   { datasystem_selection_free();
     datasystem_selection_init();

     Datasystem * _datasystem_selection = _datasystem_pointer[DATASYSTEM_SELECTION];
     Datachain * _datachain = _datachain_selection_min;
     Datachain * _datachain_sel = _datasystem_selection->_datachain_first;
     Datachain * _datachain_next;
     int delta;
     do
        { delta = _datachain->selection_end - _datachain->selection_start;
          memcopy(_datachain_sel->_line, _datachain->_line + _datachain->selection_start, delta);
          _datachain_sel->size = delta;
          _datachain_sel->invisible = _datachain->invisible;
          _datachain_next = _datachain->_next;
          if ( flag == CUT_FLAG )
             { selection_shift(_datachain);
             }
          dataline_strip(_datachain_sel);
          _datachain = _datachain_next;
          if ( _datachain != _datachain_selection_max->_next )
             { _datachain_sel = _datachain_insert(_datasystem_selection, _datachain_sel);
             }
        }
     while ( _datachain != _datachain_selection_max->_next );

     if ( flag == CUT_FLAG )
        { bool remove_rq = 1;
          _datachain = _datachain_selection_min;
          do
             { if ( ! dataline_is_empty(_datachain) )
                  { remove_rq = 0;
                    break;
                  }
               _datachain = _datachain->_next;
             }
          while ( _datachain != _datachain_selection_max->_next );
          if ( remove_rq )
             { block_remove(_datachain_selection_min, _datachain_selection_max, 0);
             }
        }
     datasystem_verify(_datasystem_selection);
     _datasystem_selection->status_register |= MODIFY_STATE;
     screen_redraw();

     return;
   }
void selection_paste()
   { if ( datasystem_ordinal == DATASYSTEM_SELECTION )
        { return;
        }
     if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }
     if ( datasystem_selection_none() )
        { status_line_print(_message[38]);
          return;
        }
     Datachain * _datachain_orig = * __datachain;
     int cursor_x = * _cursor_x;
     int dataline_offset = * _dataline_offset;
     int status_register = * _status_register;
     * _status_register |= INS_MODE;

     Datasystem * _datasystem_selection = _datasystem_pointer[DATASYSTEM_SELECTION];
     Datachain * _datachain_sel = _datasystem_selection->_datachain_first;
     do
        { for ( int i = 0; i < _datachain_sel->size; i++ )
             { dataline_fill(* (_datachain_sel->_line + i));
             }
          (* __datachain)->invisible = _datachain_sel->invisible;
          _datachain_sel = _datachain_sel->_next;
          if ( _datachain_sel )
             { * __datachain = _datachain_insert(_datasystem, * __datachain);
               * _dataline_offset = dataline_offset;
             }
        }
     while ( _datachain_sel );

     * __datachain = _datachain_orig;
     * _cursor_x = cursor_x;
     * _dataline_offset = dataline_offset;
     * _status_register = status_register;
     * _status_register |= MODIFY_STATE;
     screen_redraw();

     return;
   }

void selection_move_left()
   {     // -- check selected block for space leftwards --
     Datachain * _datachain = _datachain_selection_min;
     char * _dataline;
     do
       { _dataline = _datachain->_line;
         if ( ! _datachain->selection_start )
             { return;
             }
         if ( * (_dataline + _datachain->selection_start - 1) != ' ' )
            { return;
            }
         _datachain = _datachain->_next;
       }
     while ( _datachain != _datachain_selection_max->_next );

        // -- move selected block lefwards --
     _datachain = _datachain_selection_min;
     do
        { selection_shift_left(_datachain);
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

     selection_color();

     return;
   }
void selection_move_right()
   {     // -- check selected block for space rightwards --
     Datachain * _datachain = _datachain_selection_min;
     char * _dataline;
     do
        { _dataline = _datachain->_line;
          if ( * (_dataline + _datachain->selection_end) != ' ' )
             { return;
             }
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

        // -- move selected block rightwards --
     _datachain = _datachain_selection_min;
     do
        { selection_shift_right(_datachain);
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

     selection_color();

     return;
   }

bool selection_move_up_c3()
   {    // -- collapsed block upwards --
     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     if ( ! _datachain_prev->invisible )
        { return 0;
        }
     Datachain * _datachain_prev_new = _datachain_prev->_prev;
     if ( ! _datachain_prev_new )
        { return 0;
        }
     Datachain * _datachain_block_header = _next_visible_line_up(_datachain_selection_min);
     if ( ! _datachain_block_header )
        { return 0;
        }
     Datachain * _datachain_block_begin, * _datachain_block_end;
     int block_level = block_locate(_datachain_block_header, & _datachain_block_begin, & _datachain_block_end);
     if ( ! block_level || block_level == -1 )
        { return 0;
        }
     if ( _datachain_prev != _datachain_block_end )
        { return 0;
        }
     _datachain_prev_new = _datachain_block_header->_prev;
     if ( ! _datachain_prev_new )
        { * __datachain_first = _datachain_selection_min;
          _datachain_selection_min->_prev = 0;
        }
     else
        { _datachain_prev_new->_next = _datachain_selection_min;
          _datachain_selection_min->_prev = _datachain_prev_new;
        }
     Datachain * _datachain_next = _datachain_selection_max->_next;
     _datachain_selection_max->_next = _datachain_block_header;
     _datachain_block_header->_prev = _datachain_selection_max;
     if ( ! _datachain_next )
        { * __datachain_last = _datachain_prev;
          _datachain_prev->_next = 0;
        }
     else
        { _datachain_next->_prev = _datachain_prev;
          _datachain_prev->_next = _datachain_next;
        }

        // -- selection ordinal --
     int delta_b = _datachain_block_end->ordinal - _datachain_block_header->ordinal + 1;
     int delta_s = _datachain_selection_max->ordinal - _datachain_selection_min->ordinal + 1;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal -= delta_b;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

        // -- block ordinal --
     //int delta_s = _datachain_selection_max->ordinal - _datachain_selection_min->ordinal + 1;
     _datachain = _datachain_block_header;
     do
        { _datachain->ordinal += delta_s;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_block_end->_next );

     datasystem_verify(_datasystem);

     return 1;
   }
bool selection_move_up_c2()
   {    // -- two lines upwards, first is visible --
     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     if ( _datachain_prev->invisible )
        { return 0;
        }
     Datachain * _datachain_prev_new = _datachain_prev->_prev;
     if ( ! _datachain_prev_new )
        { return 0;
        }
     _datachain_prev_new->_next = _datachain_selection_min;
     _datachain_selection_min->_prev = _datachain_prev_new;

     Datachain * _datachain_next = _datachain_selection_max->_next;
     _datachain_selection_max->_next = _datachain_prev;
     _datachain_prev->_prev = _datachain_selection_max;
     if ( ! _datachain_next )
        { * __datachain_last = _datachain_prev;
          _datachain_prev->_next = 0;
        }
     else
        { _datachain_next->_prev = _datachain_prev;
          _datachain_prev->_next = _datachain_next;
        }
         
        // -- ordinals --
     _datachain_prev->ordinal = _datachain_selection_max->ordinal;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal--;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );
                                                             
     datasystem_verify(_datasystem);
                         
     return 1;           
   }
bool selection_move_up_c1()
   {    // -- a single visible line upwards --
     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     if ( _datachain_prev->invisible )
        { return 0;
        }
     Datachain * _datachain_prev_new = _datachain_prev->_prev;
     if ( _datachain_prev_new )
        { return 0;
        }

     * __datachain_first = _datachain_selection_min;
     _datachain_selection_min->_prev = 0;

     Datachain * _datachain_next = _datachain_selection_max->_next;
     _datachain_selection_max->_next = _datachain_prev;
     _datachain_prev->_prev = _datachain_selection_max;
     if ( ! _datachain_next )
        { * __datachain_last = _datachain_prev;
          _datachain_prev->_next = 0;
        }
     else
        { _datachain_next->_prev = _datachain_prev;
          _datachain_prev->_next = _datachain_next;
        }

        // -- ordinals --
     _datachain_prev->ordinal = _datachain_selection_max->ordinal;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal--;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

     datasystem_verify(_datasystem);

     return 1;
   }
void selection_move_up()
   { if ( selection_visible_lines() > screen_bottom - screen_top - 1 )
        { return;
        }
     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     if ( ! _datachain_prev )
        { return;
        }
     bool m = 0;
     Datachain * _datachain_prev_new = _datachain_prev->_prev;
     if ( ! _datachain_prev_new )
        { m = selection_move_up_c1();
        }
     else
        { if ( ! _datachain_prev->invisible )
             { m = selection_move_up_c2();
             }
          else
             { m = selection_move_up_c3();
             }
        }
     if ( ! m )
        { return;
        }

     * _status_register |= MODIFY_STATE;

     if ( selection_y_direction == SELECTION_UP )
        { if ( * _cursor_y > screen_top + 1 )
             { (* _cursor_y)--;
             }
          else
             { if ( visible_lines_up(_datachain_selection_min) == 1 )
                  { (* _cursor_y)--;
                  }
             }
        }
     else
        { int v = selection_visible_lines();
          if ( * _cursor_y - v + 1 > screen_top + 1 )
             { (* _cursor_y)--;
             }
          else
             { if ( visible_lines_up(_datachain_selection_min) == 1 )
                  { (* _cursor_y)--;
                  }
             }
        }
     screen_redraw();
     selection_color();

     return;
   }

bool selection_move_down_c3()
   { Datachain * _datachain_next = _datachain_selection_max->_next;
     if ( _datachain_next->invisible )
        { return 0;
        }
     Datachain * _datachain_next_new = _datachain_next->_next;
     if ( ! _datachain_next_new )
        { return 0;
        }
     Datachain * _datachain_block_header = _next_visible_line_down(_datachain_selection_max);
     if ( ! _datachain_block_header )
        { return 0;
        }
     Datachain * _datachain_block_begin, * _datachain_block_end;
     int block_level = block_locate(_datachain_block_header, & _datachain_block_begin, & _datachain_block_end);
     if ( ! block_level || block_level == -1 )
        { return 0;
        }
     if ( _datachain_next != _datachain_block_header )
        { return 0;
        }
     _datachain_next_new = _next_visible_line_down(_datachain_block_header);
     if ( ! _datachain_next_new )
        { * __datachain_last = _datachain_selection_max;
          _datachain_selection_max->_next = 0;
        }
     else
        { _datachain_next_new->_prev = _datachain_selection_max;
          _datachain_selection_max->_next = _datachain_next_new;
        }
     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     _datachain_selection_min->_prev = _datachain_block_end;
     _datachain_block_end->_next = _datachain_selection_min;
     if ( ! _datachain_prev )
        { * __datachain_first = _datachain_block_header;
          _datachain_block_header->_prev = 0;
        }
     else
        { _datachain_prev->_next = _datachain_block_header;
          _datachain_block_header->_prev = _datachain_prev;
        }

        // -- ordinal --
     int delta_b = _datachain_block_end->ordinal - _datachain_block_header->ordinal + 1;
     int delta_s = _datachain_selection_max->ordinal - _datachain_selection_min->ordinal + 1;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal += delta_b;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );
                                                             
        // -- block ordinal --
     _datachain = _datachain_block_header;
     do
        { _datachain->ordinal -= delta_s;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_block_end->_next );
                                                         
     datasystem_verify(_datasystem);
                         
     return 1;
   }
bool selection_move_down_c2()
   {   // -- two lines downwards, first is visible --
     Datachain * _datachain_next = _datachain_selection_max->_next;
     if ( _datachain_next->invisible )
        { return 0;
        }
     Datachain * _datachain_next_new = _datachain_next->_next;
     if ( ! _datachain_next_new )
        { return 0;
        }

     _datachain_next_new->_prev = _datachain_selection_max;
     _datachain_selection_max->_next = _datachain_next_new;

     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     _datachain_next->_next = _datachain_selection_min;
     _datachain_selection_min->_prev = _datachain_next;
     if ( ! _datachain_prev )
        { * __datachain_first = _datachain_next;
          _datachain_next->_prev = 0;
        }
     else
        { _datachain_prev->_next = _datachain_next;
          _datachain_next->_prev = _datachain_prev;
        }

        // -- ordinal --
     _datachain_next->ordinal = _datachain_selection_min->ordinal;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal++;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );
     //while ( _datachain );
                                                             
     datasystem_verify(_datasystem);                                    
                                                             
     return 1;                                               
   }
bool selection_move_down_c1()
   {   // -- a single visible line downwards --
     Datachain * _datachain_next = _datachain_selection_max->_next;
     if ( _datachain_next->invisible )
        { return 0;
        }
     Datachain * _datachain_next_new = _datachain_next->_next;
     if ( _datachain_next_new )
        { return 0;
        }

     * __datachain_last = _datachain_selection_max;
     _datachain_selection_max->_next = 0;

     Datachain * _datachain_prev = _datachain_selection_min->_prev;
     _datachain_next->_next = _datachain_selection_min;
     _datachain_selection_min->_prev = _datachain_next;
     if ( ! _datachain_prev )
        { * __datachain_first = _datachain_next;
          _datachain_next->_prev = 0;
        }
     else
        { _datachain_prev->_next = _datachain_next;
          _datachain_next->_prev = _datachain_prev;
        }

        // -- ordinal --
     _datachain_next->ordinal = _datachain_selection_min->ordinal;
     Datachain * _datachain = _datachain_selection_min;
     do
        { _datachain->ordinal++;
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );
     //while ( _datachain );
                                                             
     datasystem_verify(_datasystem);                                    
                                                             
     return 1;                                               
   }
void selection_move_down()
   { if ( selection_visible_lines() > screen_bottom - screen_top - 1 )
        { return;
        }
     Datachain * _datachain_next = _datachain_selection_max->_next;
     if ( ! _datachain_next )
        { return;
        }
     bool m = 0;
     Datachain * _datachain_next_new = _datachain_next->_next;
     if ( ! _datachain_next_new )
        { m = selection_move_down_c1();
        }
     else
        { if ( ! _datachain_next_new->invisible )
             { m = selection_move_down_c2();
             }
          else
             { m = selection_move_down_c3();
             }
        }
     if ( ! m )
        { return;
        }

     * _status_register |= MODIFY_STATE;

     if ( selection_y_direction == SELECTION_DOWN )
        { if ( * _cursor_y < screen_bottom - 1 )
             { (* _cursor_y)++;
             }
          else
             { if ( visible_lines_down(_datachain_selection_max) == 1 )
                  { (* _cursor_y)++;
                  }
             }
        }
     else
        { int v = selection_visible_lines();
          if ( * _cursor_y + v - 1 < screen_bottom - 1 )
             { (* _cursor_y)++;
             }
          else
             { if ( visible_lines_down(_datachain_selection_max) == 1 )
                  { (* _cursor_y)++;
                  }
             }
        }
     screen_redraw();
     selection_color();
            
     return;
   }

void selection_clear()
   { Datachain * _datachain = * __datachain_first;

     do
        { _datachain->selection_start = 0;
          _datachain->selection_end = 0;
          _datachain = _datachain->_next;
        }
     while ( _datachain );

     return;
   }

bool datachain_is_selected(Datachain * _datachain)
   { if ( ! _datachain->selection_start && ! _datachain->selection_end )
        { return 0;
        }
     char * _dataline_point = _dataline_advance(_datachain, _datachain->_line);
     if ( _datachain->selection_start > _dataline_point - _datachain->_line )
        { return 0;
        }
     if ( _datachain->selection_end < _datachain->size )
        { return 0;
        }
     return 1;
   }
bool block_select(Datachain * _datachain_header, Datachain ** __datachain_block_begin, Datachain ** __datachain_block_end)
   { int block_level = block_locate(_datachain_header, __datachain_block_begin, __datachain_block_end);
     if ( ! block_level || block_level == -1 )
        { return 0;
        }

     Datachain * _datachain = * __datachain_block_begin;
     do
        { _datachain->selection_start = selection_start;
          _datachain->selection_end = _datachain->size;
          _datachain = _datachain->_next;
        }
     while ( _datachain != (* __datachain_block_end)->_next);

     return 1;
   }

int selection_visible_lines()
   { selection_update();

     Datachain * _datachain = _datachain_selection_min;
     int s = 0;
     do
        { if ( ! _datachain->invisible )
             { s++;
             }
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

     return s;
   }
int selection_line_max()
   { int line_max = 0;
     int delta;

     Datachain * _datachain = _datachain_selection_min;
     do
        { delta = _datachain->selection_end - _datachain->selection_start;
          if ( delta > line_max )
             { line_max = delta;
             }
          _datachain = _datachain->_next;
        }
     while ( _datachain != _datachain_selection_max->_next );

     return line_max;
   }

void selection_view_in()
   { if ( datasystem_selection_none() )
        { status_line_print(_message[38]);
          return;
        }
     datasystem_ordinal_main = datasystem_ordinal;
     datasystem_ordinal = DATASYSTEM_SELECTION;
     datasystem_set();
     header_line_refresh();
     screen_redraw();

     return;
   }
void selection_view_out()
   { datasystem_ordinal = datasystem_ordinal_main;
     datasystem_set();
     header_line_refresh();
     screen_redraw();

     return;
   }

void dataline_strip(Datachain * _datachain)
   { if ( ! _datachain->size )
        { return;
        }
     char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline + _datachain->size - 1;
     while ( _dataline_point != _dataline )
        { if ( * _dataline_point-- == ' ' )
             { _datachain->size--;
             }
          else
             { break;
             }
        }
     return;
   }
